/*
 * REX-5052 GPIB driver - 1.2
 *
 * Copyright (C) 2005 RatocSystems, Inc.
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 *
 * Based on dummy PCMCIA client driver
 *
 * Changelog:
 * 2005/12/01  start coding
 * 2006/12/13  changed  interruptible_sleep_on()->wait_event_interruptible()
  *
 **/

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/version.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <linux/fs.h>		/* for struct file_operations */
#include <linux/wait.h>
#include <linux/interrupt.h>

#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/cisreg.h>
#include <pcmcia/ds.h>

#include "rex5052.h"


#define MY_MAJOR 160
#define MY_MINOR 0
#define MY_TIMEOUT	(2*HZ)			/* 2 seconds */


/* Structure to hold all of our device specific stuff */
struct rs_gpib {
	int	iobase;
	int	irq;
	int	enable;
	wait_queue_head_t waitq;
	int	waitq_complete;
	struct semaphore lock;          /* general race avoidance */
};

struct rs_gpib gpib_dev_tbl[4];


/*====================================================================*/

/* Module parameters */

#define INT_MODULE_PARM(n, v) static int n = v; MODULE_PARM(n, "i")

/* Release IO ports after configuration? */
//INT_MODULE_PARM(free_ports, 0);
static int free_ports = 0;
static int major;
static struct class *gpib_class;

/*
   All the PCMCIA modules use PCMCIA_DEBUG to control debugging.  If
   you do not define PCMCIA_DEBUG at all, all the debug code will be
   left out.  If you compile with PCMCIA_DEBUG=0, the debug code will
   be present but disabled -- but it can then be enabled for specific
   modules at load time with a 'pc_debug=#' option to insmod.
*/

#ifdef PCMCIA_DEBUG
//INT_MODULE_PARM(pc_debug, PCMCIA_DEBUG);
//#define DEBUG(n, args...) if (pc_debug>(n)) printk(KERN_DEBUG args)
//#define DEBUG(n, args...) printk(KERN_DEBUG args)
static char *version = "ver 1.2";
#else
#define DEBUG(n, args...)
#endif

/*====================================================================*/

/*
   You'll also need to prototype all the functions that will actually
   be used to talk to your device.  See 'memory_cs' for a good example
   of a fully self-sufficient driver; the other drivers rely more or
   less on other parts of the kernel.
*/

/*
   The dev_info variable is the "key" that is used to match up this
   device driver with appropriate cards, through the card configuration
   database.
*/

static dev_info_t dev_info = "rex5052_cs";

/*
   A linked list of "instances" of the dummy device.  Each actual
   PCMCIA card corresponds to one device instance, and is described
   by one dev_link_t structure (defined in ds.h).

   You may not want to use a linked list for this -- for example, the
   memory card driver uses an array of dev_link_t pointers, where minor
   device numbers are used to derive the corresponding array index.
*/

static dev_link_t *dev_list = NULL;

/*
   A dev_link_t structure has fields for most things that are needed
   to keep track of a socket, but there will usually be some device
   specific information that also needs to be kept track of.  The
   'priv' pointer in a dev_link_t structure can be used to point to
   a device-specific private data structure, like this.

   To simplify the data structure handling, we actually include the
   dev_link_t structure in the device's private data structure.

   A driver needs to provide a dev_node_t structure for each device
   on a card.  In some cases, there is only one device per card (for
   example, ethernet cards, modems).  In other cases, there may be
   many actual or logical devices (SCSI adapters, memory cards with
   multiple partitions).  The dev_node_t structures need to be kept
   in a linked list starting at the 'dev' field of a dev_link_t
   structure.  We allocate them in the card's private data structure,
   because they generally shouldn't be allocated dynamically.

   In this case, we also provide a flag to indicate if a device is
   "stopped" due to a power management event, or card ejection.  The
   device IO routines can use a flag like this to throttle IO to a
   card that is not ready to accept it.

   The bus_operations pointer is used on platforms for which we need
   to use special socket-specific versions of normal IO primitives
   (inb, outb, readb, writeb, etc) for card IO.
*/
   
typedef struct local_info_t {
    dev_link_t		link;
    dev_node_t		node;
    int			stop;
    struct bus_operations *bus;
} local_info_t;


static void rs_gpib_config(dev_link_t * link);
static void rs_gpib_cs_release(dev_link_t *);

/*======================================================================

    rs_gpib_attach() creates an "instance" of the driver, allocating
    local data structures for one device.  The device is registered
    with Card Services.

======================================================================*/
static int rs_gpib_attach(struct pcmcia_device *p_dev)
{
    local_info_t *info;
    dev_link_t *link;

#ifdef PCMCIA_DEBUG
    printk(KERN_INFO "rs_gpib_attach()\n");
#endif

    /* Create new parport device */
    info = kmalloc(sizeof(*info), GFP_KERNEL);
    if (!info) return -ENOMEM;
    memset(info, 0, sizeof(*info));
    link = &info->link; link->priv = info;

    link->io.Attributes1 = IO_DATA_PATH_WIDTH_8;
    link->io.Attributes2 = IO_DATA_PATH_WIDTH_8;
    link->irq.Attributes = IRQ_TYPE_EXCLUSIVE;
    link->irq.IRQInfo1 = IRQ_LEVEL_ID;
    link->conf.Attributes = CONF_ENABLE_IRQ;
    link->conf.Vcc = 50;
    link->conf.IntType = INT_MEMORY_AND_IO;

    link->next = dev_list;
    dev_list = link;

    link->handle = p_dev;
    p_dev->instance = link;

    link->state |= DEV_PRESENT | DEV_CONFIG_PENDING;
    rs_gpib_config(link);

	class_device_create(gpib_class, NULL, MKDEV(major, 0), NULL,
			    "gpib%d", 0);

    return 0;
} /* rs_gpib_attach */

/*======================================================================

    This deletes a driver "instance".  The device is de-registered
    with Card Services.  If it has been released, all local data
    structures are freed.  Otherwise, the structures will be freed
    when the device is released.

======================================================================*/
static void rs_gpib_detach(struct pcmcia_device *p_dev)
{
    dev_link_t *link = dev_to_instance(p_dev);

#ifdef PCMCIA_DEBUG
    printk(KERN_INFO "rs_gpib_detach(0x%p)\n", link);
#endif

    if (link->state & DEV_CONFIG)
		rs_gpib_cs_release(link);

	link->state &= ~DEV_PRESENT;

    /* This points to the parent local_info_t struct */
    kfree(link->priv);

	class_device_destroy(gpib_class, MKDEV(major, 0));
    
} /* rs_gpib_detach */

/*======================================================================

    rs_gpib_config() is scheduled to run after a CARD_INSERTION event
    is received, to configure the PCMCIA socket, and to make the
    device available to the system.

======================================================================*/

static void rs_gpib_config(dev_link_t *link)
{
    local_info_t *local = link->priv;
    tuple_t tuple;
    cisparse_t parse;
    int last_fn = 0;
    int last_ret;
    u_char buf[64];
    config_info_t conf;
    win_req_t req;
    memreq_t map;
    cistpl_cftable_entry_t dflt = { 0 };
    //struct rs_gpib *p = NULL;

    /*
       This reads the card's CONFIG tuple to find its configuration
       registers.
    */
    tuple.DesiredTuple = CISTPL_CONFIG;
    tuple.Attributes = 0;
    tuple.TupleData = buf;
    tuple.TupleDataMax = sizeof(buf);
    tuple.TupleOffset = 0;
	last_ret = pcmcia_get_first_tuple(link->handle, &tuple);
	if(last_ret != CS_SUCCESS) 
	{
		last_fn = GetFirstTuple;
		goto cs_failed;
	}
	last_ret = pcmcia_get_tuple_data(link->handle, &tuple);
	if(last_ret != CS_SUCCESS) 
	{
		last_fn = GetTupleData;
		goto cs_failed;
	}
	last_ret = pcmcia_parse_tuple(link->handle, &tuple, &parse);
	if(last_ret != CS_SUCCESS) 
	{
		last_fn = ParseTuple;
		goto cs_failed;
	}
    link->conf.ConfigBase = parse.config.base;
    link->conf.Present = parse.config.rmask[0];

    /* Configure card */
    link->state |= DEV_CONFIG;

    /* Look up the current Vcc */
	last_ret = pcmcia_get_configuration_info(link->handle, &conf);
	if(last_ret != CS_SUCCESS) 
	{
		last_fn = GetConfigurationInfo;
		goto cs_failed;
	}
    link->conf.Vcc = conf.Vcc;

    /*
      In this loop, we scan the CIS for configuration table entries,
      each of which describes a valid card configuration, including
      voltage, IO window, memory window, and interrupt settings.

      We make no assumptions about the card to be configured: we use
      just the information available in the CIS.  In an ideal world,
      this would work for any PCMCIA card, but it requires a complete
      and accurate CIS.  In practice, a driver usually "knows" most of
      these things without consulting the CIS, and most client drivers
      will only use the CIS to fill in implementation-defined details.
    */
    tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;
	last_ret = pcmcia_get_first_tuple(link->handle, &tuple);
	if(last_ret != CS_SUCCESS) 
	{
		last_fn = GetFirstTuple;
		goto cs_failed;
	}
    while (1) 
	{
		cistpl_cftable_entry_t *cfg = &(parse.cftable_entry);
		last_ret = pcmcia_get_tuple_data(link->handle, &tuple);
		if(last_ret != CS_SUCCESS) 
		{
			goto next_entry;
		}
		last_ret = pcmcia_parse_tuple(link->handle, &tuple, &parse);
		if(last_ret != CS_SUCCESS) 
		{
			goto next_entry;
		}
		if (cfg->flags & CISTPL_CFTABLE_DEFAULT) dflt = *cfg;
		if (cfg->index == 0) goto next_entry;
		link->conf.ConfigIndex = cfg->index;

		/* Does this card need audio output? */
		if (cfg->flags & CISTPL_CFTABLE_AUDIO) 
		{
			link->conf.Attributes |= CONF_ENABLE_SPKR;
			link->conf.Status = CCSR_AUDIO_ENA;
		}

		/* Use power settings for Vcc and Vpp if present */
		/*  Note that the CIS values need to be rescaled */
		if (cfg->vcc.present & (1<<CISTPL_POWER_VNOM)) 
		{
			if (conf.Vcc != cfg->vcc.param[CISTPL_POWER_VNOM]/10000)
				goto next_entry;
		}else if (dflt.vcc.present & (1<<CISTPL_POWER_VNOM)) 
		{
			if (conf.Vcc != dflt.vcc.param[CISTPL_POWER_VNOM]/10000)
				goto next_entry;
		}

		if (cfg->vpp1.present & (1<<CISTPL_POWER_VNOM))
			link->conf.Vpp1 = link->conf.Vpp2 =
				cfg->vpp1.param[CISTPL_POWER_VNOM]/10000;
		else if (dflt.vpp1.present & (1<<CISTPL_POWER_VNOM))
			link->conf.Vpp1 = link->conf.Vpp2 =
				dflt.vpp1.param[CISTPL_POWER_VNOM]/10000;
	
		/* Do we need to allocate an interrupt? */
		if (cfg->irq.IRQInfo1 || dflt.irq.IRQInfo1)
			link->conf.Attributes |= CONF_ENABLE_IRQ;
	
		/* IO window settings */
		link->io.NumPorts1 = link->io.NumPorts2 = 0;
		if ((cfg->io.nwin > 0) || (dflt.io.nwin > 0)) 
		{
			cistpl_io_t *io = (cfg->io.nwin) ? &cfg->io : &dflt.io;
			link->io.Attributes1 = IO_DATA_PATH_WIDTH_AUTO;
			if (!(io->flags & CISTPL_IO_8BIT))
				link->io.Attributes1 = IO_DATA_PATH_WIDTH_16;
			if (!(io->flags & CISTPL_IO_16BIT))
				link->io.Attributes1 = IO_DATA_PATH_WIDTH_8;
			link->io.IOAddrLines = io->flags & CISTPL_IO_LINES_MASK;
			link->io.BasePort1 = io->win[0].base;
			link->io.NumPorts1 = io->win[0].len;
			if (io->nwin > 1) 
			{
				link->io.Attributes2 = link->io.Attributes1;
				link->io.BasePort2 = io->win[1].base;
				link->io.NumPorts2 = io->win[1].len;
			}
			/* This reserves IO space but doesn't actually enable it */
			last_ret = pcmcia_request_io(link->handle, &link->io);
			if(last_ret != CS_SUCCESS) 
			{
				goto next_entry;
			}
		}
	
		/*
		Now set up a common memory window, if needed.  There is room
		in the dev_link_t structure for one memory window handle,
		but if the base addresses need to be saved, or if multiple
		windows are needed, the info should go in the private data
		structure for this device.
	
		Note that the memory window base is a physical address, and
		needs to be mapped to virtual space with ioremap() before it
		is used.
		*/
		if ((cfg->mem.nwin > 0) || (dflt.mem.nwin > 0)) {
			cistpl_mem_t *mem =
			(cfg->mem.nwin) ? &cfg->mem : &dflt.mem;
			req.Attributes = WIN_DATA_WIDTH_16|WIN_MEMORY_TYPE_CM;
			req.Attributes |= WIN_ENABLE;
			req.Base = mem->win[0].host_addr;
			req.Size = mem->win[0].len;
			if (req.Size < 0x1000)
			req.Size = 0x1000;
			req.AccessSpeed = 0;
			link->win = (window_handle_t)link->handle;
			if(pcmcia_request_window(&link->handle, &req, &link->win) != CS_SUCCESS) goto next_entry;
			map.Page = 0; map.CardOffset = mem->win[0].card_addr;
			if(pcmcia_map_mem_page(link->win, &map) != CS_SUCCESS) goto next_entry;
		}
		/* If we got this far, we're cool! */
		break;
	
		next_entry:
		if (link->io.NumPorts1)
			pcmcia_release_io(link->handle, &link->io);
		last_ret = pcmcia_get_next_tuple(link->handle, &tuple);
		if(last_ret != CS_SUCCESS) 
		{
			last_fn = GetNextTuple;
			goto cs_failed;
		}
	}

    /*
       Allocate an interrupt line.  Note that this does not assign a
       handler to the interrupt, unless the 'Handler' member of the
       irq structure is initialized.
    */
    if (link->conf.Attributes & CONF_ENABLE_IRQ)
	last_ret = pcmcia_request_irq(link->handle, &link->irq);
	if(last_ret != CS_SUCCESS) 
	{
		last_fn = RequestIRQ;
		goto cs_failed;
	}

    /*
       This actually configures the PCMCIA socket -- setting up
       the I/O windows and the interrupt mapping, and putting the
       card and host interface into "Memory and IO" mode.
    */
	last_ret = pcmcia_request_configuration(link->handle, &link->conf);
	if(last_ret != CS_SUCCESS) 
	{
		last_fn = RequestConfiguration;
		goto cs_failed;
	}

    /*
      We can release the IO port allocations here, if some other
      driver for the card is going to loaded, and will expect the
      ports to be available.
    */
	if (free_ports) {
		if (link->io.BasePort1)
			release_region(link->io.BasePort1, link->io.NumPorts1);
		if (link->io.BasePort2)
			release_region(link->io.BasePort2, link->io.NumPorts2);
	}

    /*
      At this point, the dev_node_t structure(s) need to be
      initialized and arranged in a linked list at link->dev.
    */
    local->node.major = major;
    local->node.minor = 0;
    sprintf(local->node.dev_name, "gpib0");
    link->dev = &local->node;

#ifdef PCMCIA_DEBUG
    /* Finally, report what we've done */
    printk(KERN_INFO "%s: index 0x%02x: Vcc %d.%d",
	   local->node.dev_name, link->conf.ConfigIndex,
	   link->conf.Vcc/10, link->conf.Vcc%10);
	if (link->conf.Vpp1)
		printk(", Vpp %d.%d", link->conf.Vpp1/10, link->conf.Vpp1%10);
	if (link->conf.Attributes & CONF_ENABLE_IRQ)
		printk(", irq %d", link->irq.AssignedIRQ);
	if (link->io.NumPorts1)
		printk(", io 0x%04x-0x%04x", link->io.BasePort1,
	       link->io.BasePort1+link->io.NumPorts1-1);
	if (link->io.NumPorts2)
		printk(" & 0x%04x-0x%04x", link->io.BasePort2,
	       link->io.BasePort2+link->io.NumPorts2-1);
	if (link->win)
		printk(", mem 0x%06lx-0x%06lx", req.Base,
	       req.Base+req.Size-1);
    printk("\n");
#endif

    link->state &= ~DEV_CONFIG_PENDING;
    return;

cs_failed:
    cs_error(link->handle, last_fn, last_ret);
    rs_gpib_cs_release(link);

} /* rs_gpib_config */

void rs_gpib_cs_release(dev_link_t *link)
{
    local_info_t *info = link->priv;
    
#ifdef PCMCIA_DEBUG
	printk(KERN_INFO "rs_gpib_cs_release(0x%p)\n", link);
#endif

    {
		request_region(link->io.BasePort1, link->io.NumPorts1,
		       info->node.dev_name);
		if (link->io.NumPorts2)
		    request_region(link->io.BasePort2, link->io.NumPorts2,
			   info->node.dev_name);
    }
    link->dev = NULL;
    
    pcmcia_release_configuration(link->handle);
    pcmcia_release_io(link->handle, &link->io);
    pcmcia_release_irq(link->handle, &link->irq);
    
    link->state &= ~DEV_CONFIG;

} /* rs_gpib_cs_release */

static int rs_gpib_suspend(struct pcmcia_device *dev)
{
	dev_link_t *link = dev_to_instance(dev);
	link->state |= DEV_SUSPEND;

	if (link->state & DEV_CONFIG)
		pcmcia_release_configuration(link->handle);

	return 0;
}

static int rs_gpib_resume(struct pcmcia_device *dev)
{
    dev_link_t *link = dev_to_instance(dev);
    link->state &= ~DEV_SUSPEND;

    if (DEV_OK(link))
        pcmcia_request_configuration(link->handle, &link->conf);

    return 0;
}
static struct pcmcia_device_id rs_gpib_ids[] = {
	// # /sbin/pccardctl info
	PCMCIA_DEVICE_MANF_CARD(0x0000, 0x0001),
	PCMCIA_DEVICE_PROD_ID12("RATOC System,Inc.", "REX5052 GPIB PC Card", 0x4031e316, 0x7eedf861),
	PCMCIA_DEVICE_NULL
};
MODULE_DEVICE_TABLE(pcmcia, rs_gpib_ids);

static struct pcmcia_driver rex5052_cs_driver = {
	.owner		= THIS_MODULE,
	.drv		= {
		.name	= "rex5052_cs",
	},
	.probe		= rs_gpib_attach,
	.remove		= rs_gpib_detach,
	.id_table	= rs_gpib_ids,
	.suspend	= rs_gpib_suspend,
	.resume		= rs_gpib_resume,
};

int __init init_rex5052_cs(void)
{
#ifdef PCMCIA_DEBUG
	printk(KERN_INFO "init_rex5052_cs: %s\n", version);
#endif
	gpib_class = class_create(THIS_MODULE, "rex5052_gpib");
	if (!gpib_class)
		return -1;

	pcmcia_register_driver(&rex5052_cs_driver);
	return 0;
}

void __exit exit_rex5052_cs(void)
{
#ifdef PCMCIA_DEBUG
	printk(KERN_INFO "exit_rex5052_cs: unloading\n");
#endif
	pcmcia_unregister_driver(&rex5052_cs_driver);
	class_destroy(gpib_class);
}

//module_init(init_rex5052_cs);
//module_exit(exit_rex5052_cs);


// interrupt handler
void gpib_interrupt(int irq, void *dev_instance, struct pt_regs *regs)
{
	struct rs_gpib *dev = (struct rs_gpib *)dev_instance;
#if 0
//#ifdef PCMCIA_DEBUG
	char	isr0, isr1;

	// read Interrupt Status Register 0
	isr0 = inb(dev->iobase+0);
	// read Interrupt Status Register 1
	isr1 = inb(dev->iobase+2);
	// ex) isr0:0x50 isr1:0x02
	printk(KERN_INFO "gpib_interrupt irq:%d, isr0:0x%x, isr1:0x%x\n", irq, isr0, isr1);
#endif

	// TMS9914 interrupt disable
	outb(0, dev->iobase+2);
	// REX5052 interrupt disable
	outb(0, dev->iobase+4);
	dev->waitq_complete = 1;
	wake_up_interruptible(&dev->waitq);
}

ssize_t gpib_read(struct file *file, char *buffer, size_t count, loff_t *ppos)
{
	struct rs_gpib *dev;
	int retval = 0;

#ifdef PCMCIA_DEBUG
	printk(KERN_INFO "gpib_read count = 0x%x\n", count);
#endif
	dev = (struct rs_gpib *)file->private_data;

	return retval;
}

ssize_t gpib_write(struct file *file, const char *user_buffer, size_t count, loff_t *ppos)
{
	struct rs_gpib *dev;
	int retval = 0;

#ifdef PCMCIA_DEBUG
	printk(KERN_INFO "gpib_write count = 0x%x\n", count);
#endif
	dev = (struct rs_gpib *)file->private_data;

	return retval;
}

int gpib_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	struct rs_gpib *dev;
	CARD_RESOURCE		card;
	IOPORT_UCHAR		bIOPort;
	IOPORT_USHORT		wIOPort;
	void __user *data;
	int retval = 0;

//#ifdef PCMCIA_DEBUG
//	printk(KERN_INFO "gpib_ioctl cmd = 0x%x\n", cmd);
//#endif
	dev = (struct rs_gpib *)file->private_data;
	down(&(dev->lock));
	/* Sanity check to make sure rio is connected, powered, etc */
    if ( dev == NULL)
	{
		retval = -ENODEV;
		goto err_out;
	}

	switch (cmd) {
	case IOCTL_GP_GET_RESOURCE:
		data = (void __user *) arg;
		if (data == NULL)
			break;
		card.IOBaseLow = dev->iobase;
		card.IOBaseHigh = dev->iobase+dev_list->io.NumPorts1-1;
		card.IOLength = dev_list->io.NumPorts1;
		card.IrqLevel = dev->irq;
		if (copy_to_user(data, &card, sizeof(CARD_RESOURCE))) {
			retval = -EFAULT;
			goto err_out;
		}
		break;
	case IOCTL_GP_READ_PORT_UCHAR:
		data = (void __user *) arg;
		if (data == NULL)
			break;
		if (copy_from_user(&bIOPort, data, sizeof(IOPORT_UCHAR))) {
			retval = -EFAULT;
			goto err_out;
		}
		bIOPort.Value = inb(dev->iobase+bIOPort.Offset);
		if (copy_to_user(data, &bIOPort, sizeof(IOPORT_UCHAR))) {
			retval = -EFAULT;
			goto err_out;
		}
//#ifdef PCMCIA_DEBUG
//		printk(KERN_INFO "inb: offset = 0x%x value = 0x%x\n", bIOPort.Offset, bIOPort.Value);
//#endif
		break;
	case IOCTL_GP_READ_PORT_USHORT:
		data = (void __user *) arg;
		if (data == NULL)
			break;
		if (copy_from_user(&wIOPort, data, sizeof(IOPORT_USHORT))) {
			retval = -EFAULT;
			goto err_out;
		}
		wIOPort.Value = inw(dev->iobase+wIOPort.Offset);
		if (copy_to_user(data, &wIOPort, sizeof(IOPORT_USHORT))) {
			retval = -EFAULT;
			goto err_out;
		}
#ifdef PCMCIA_DEBUG
		printk(KERN_INFO "inw: offset = 0x%x value = 0x%x\n", wIOPort.Offset, wIOPort.Value);
#endif
		break;
	case IOCTL_GP_WRITE_PORT_UCHAR:
		data = (void __user *) arg;
		if (data == NULL)
			break;
		if (copy_from_user(&bIOPort, data, sizeof(IOPORT_UCHAR))) {
			retval = -EFAULT;
			goto err_out;
		}
		outb(bIOPort.Value, dev->iobase+bIOPort.Offset);
//#ifdef PCMCIA_DEBUG
//		printk(KERN_INFO "outb offset 0x%x value = 0x%x\n", 
//				bIOPort.Offset, bIOPort.Value);
//#endif
		break;
	case IOCTL_GP_WRITE_PORT_USHORT:
		data = (void __user *) arg;
		if (data == NULL)
			break;
		if (copy_from_user(&wIOPort, data, sizeof(IOPORT_USHORT))) {
			retval = -EFAULT;
			goto err_out;
		}
		outw(wIOPort.Value, dev->iobase+wIOPort.Offset);
#ifdef PCMCIA_DEBUG
		printk(KERN_INFO "outw offset 0x%x value = 0x%x\n", 
				wIOPort.Offset, wIOPort.Value);
#endif
		break;
	case IOCTL_GP_START_INTERRUPT:
#ifdef PCMCIA_DEBUG
		printk(KERN_INFO "gpib_ioctl: IOCTL_GP_START_INTERRUPT\n");
		printk(KERN_INFO "gpib_ioctl INTERRUPT...->sleep_on\n");
#endif
		up(&(dev->lock));
		//interruptible_sleep_on(&dev->waitq);
		dev->waitq_complete = 0;
		wait_event_interruptible(dev->waitq, (dev->waitq_complete) );
		down(&(dev->lock));
#ifdef PCMCIA_DEBUG
		printk(KERN_INFO "gpib_ioctl INTERRUPT...sleep_on->\n");
#endif
		break;
	case IOCTL_GP_STOP_INTERRUPT:
#ifdef PCMCIA_DEBUG
		printk(KERN_INFO "gpib_ioctl: IOCTL_GP_STOP_INTERRUPT\n");
#endif
		// TMS9914 interrupt disable
		outb(0, dev->iobase+2);
		// REX5052 interrupt disable
		outb(0, dev->iobase+4);
		dev->waitq_complete = 1;
		wake_up_interruptible(&dev->waitq);
		break;
	default:
		retval = -ENOIOCTLCMD;
		break;
	}

err_out:
	up(&(dev->lock));
	return retval;
}

int gpib_open(struct inode *inode, struct file *file)
{
	struct rs_gpib *dev;
	int minor;
	int retval = 0;
	int isr_flags = SA_SHIRQ;

	minor = iminor(inode);

	dev = (struct rs_gpib *)&gpib_dev_tbl[minor];
	if(dev->enable == 1){
		return	-ENODEV;
	}
	dev->enable = 1;
	dev->iobase = dev_list->io.BasePort1;
	dev->irq = dev_list->irq.AssignedIRQ;
	init_waitqueue_head(&dev->waitq);
	init_MUTEX(&(dev->lock));

	// get irq
	if( request_irq( dev_list->irq.AssignedIRQ, 
		(void *)gpib_interrupt, isr_flags, "rex5052", dev) )
	{
#ifdef PCMCIA_DEBUG
		printk(KERN_INFO "gpib_open: can't request IRQ %d\n", dev_list->irq.AssignedIRQ);
#endif
		return -1;
	}

	/* save our object in the file's private structure */
	file->private_data = dev;

#ifdef PCMCIA_DEBUG
	printk(KERN_INFO "gpib_open version = %s minor = %d\n", version, minor );
	printk(KERN_INFO "irq = %d, io 0x%04x-0x%04x\n", dev->irq, dev->iobase,
       dev->iobase+dev_list->io.NumPorts1-1);
#endif

//exit:
	return retval;
}

int gpib_release(struct inode *inode, struct file *file)
{
	struct rs_gpib *dev;

#ifdef PCMCIA_DEBUG
	printk(KERN_INFO "gpib_release subminor = %d\n", iminor(inode));
#endif
	dev = (struct rs_gpib *)file->private_data;

	if(dev == NULL || dev->enable == 0){
		return	-ENODEV;
	}
	if(dev->enable == 1){
		dev->enable = 0;
		free_irq(dev->irq, dev);
	}

	return 0;
}

static struct file_operations rex5052_fops = {
	.owner =	THIS_MODULE,
	.read =		gpib_read,
	.write =	gpib_write,
	.ioctl =	gpib_ioctl,
	.open =		gpib_open,
	.release =	gpib_release,
};

static int __init rex5052_init_module(void)
{

	/* register this driver*/
	major = register_chrdev(0, "gpib", &rex5052_fops);
	if (major < 0) {
#ifdef PCMCIA_DEBUG
		printk(KERN_INFO "register_chrdev failed. Error number %d", major);
#endif
	}
#ifdef PCMCIA_DEBUG
	printk(KERN_INFO "rex5052_init_module: loading... major %d\n", major);
#endif

	if( init_rex5052_cs() < 0 )
		return -1;

	return 0;
}

static void __exit rex5052_exit_module(void)
{
#ifdef PCMCIA_DEBUG
	printk(KERN_INFO "rex5052_exit_module: unloading\n");
#endif

	/* unregister this driver */
	unregister_chrdev(major, "gpib");

	exit_rex5052_cs();
}

module_init (rex5052_init_module);
module_exit (rex5052_exit_module);
MODULE_LICENSE("Dual BSD/GPL");
