/*
 * REX-5052 GPIB driver - 1.4
 *
 * Copyright (C) 2005-2008 RatocSystems, Inc.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation, version 2.
 *
 * Based on dummy PCMCIA client driver
 *
 * Changelog:
 * 2005/12/01  start coding
 * 2006/12/13  changed  interruptible_sleep_on()->wait_event_interruptible()
 * 2007/01
 *  - support 2.6.13 "new style" pcmcia interface
 *  - add class interface for udev device creation
 * 2008/08/18
 *  - modified interrupt routine
 *
 * All rights reserved, Dual BSD/GPL Licensed.
 *
 **/

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/version.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <linux/fs.h>		/* for struct file_operations */
#include <linux/wait.h>
#include <linux/interrupt.h>

#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/cisreg.h>
#include <pcmcia/ds.h>

#include "rex5052.h"
#include "def9914.h"

#define MY_MAJOR 160
#define MY_MINOR 0
#define MY_TIMEOUT	(2*HZ)			/* 2 seconds */
#define CM_MAX_DEV 1


/* Structure to hold all of our device specific stuff */
struct rs_gpib {
	struct pcmcia_device	*p_dev;
    dev_node_t		node;
	wait_queue_head_t waitq;
	int	waitq_complete;
	struct semaphore lock;          /* general race avoidance */
};

static struct pcmcia_device *dev_table[CM_MAX_DEV];
//static struct pcmcia_device *dev_list = NULL;

/*====================================================================*/

/* Module parameters */

static int major;
static struct class *gpib_class;

/*
   All the PCMCIA modules use PCMCIA_DEBUG to control debugging.  If
   you do not define PCMCIA_DEBUG at all, all the debug code will be
   left out.  If you compile with PCMCIA_DEBUG=0, the debug code will
   be present but disabled -- but it can then be enabled for specific
   modules at load time with a 'pc_debug=#' option to insmod.
*/

#ifdef PCMCIA_DEBUG
//INT_MODULE_PARM(pc_debug, PCMCIA_DEBUG);
//#define DEBUG(n, args...) if (pc_debug>(n)) printk(KERN_DEBUG args)
//#define DEBUG(n, args...) printk(KERN_DEBUG args)
static char *version = "ver 1.4";
#else
#define DEBUG(n, args...)
#endif


static void rs_gpib_config(struct pcmcia_device * link);
static void rs_gpib_cs_release(struct pcmcia_device *);

/*======================================================================

    rs_gpib_attach() creates an "instance" of the driver, allocating
    local data structures for one device.  The device is registered
    with Card Services.

======================================================================*/
static int rs_gpib_attach(struct pcmcia_device *link)
{
	struct rs_gpib *dev;
	int i;

#ifdef PCMCIA_DEBUG
    printk(KERN_INFO "rs_gpib_attach()\n");
#endif

	for (i = 0; i < CM_MAX_DEV; i++) {
		if (dev_table[i] == NULL)
			break;
	}

	if (i == CM_MAX_DEV)
		return -ENODEV;

    /* Create new parport device */
	dev = kzalloc(sizeof(struct rs_gpib), GFP_KERNEL);
	if (dev == NULL)
		return -ENOMEM;
    memset(dev, 0, sizeof(*dev));
	link->priv = dev;
	dev->p_dev = link;

    link->io.Attributes1 = IO_DATA_PATH_WIDTH_8;
    link->io.Attributes2 = IO_DATA_PATH_WIDTH_8;
    link->irq.Attributes = IRQ_TYPE_EXCLUSIVE;
    link->irq.IRQInfo1 = IRQ_LEVEL_ID;
    link->conf.Attributes = CONF_ENABLE_IRQ;
    link->conf.IntType = INT_MEMORY_AND_IO;

	dev_table[i] = link;
    //dev_list = link;

    rs_gpib_config(link);

	class_device_create(gpib_class, NULL, MKDEV(major, i), NULL,
			    "gpib%d", 0);

    return 0;
} /* rs_gpib_attach */

/*======================================================================

    This deletes a driver "instance".  The device is de-registered
    with Card Services.  If it has been released, all local data
    structures are freed.  Otherwise, the structures will be freed
    when the device is released.

======================================================================*/
static void rs_gpib_detach(struct pcmcia_device *link)
{
	struct rs_gpib *dev = link->priv;
	int devno;

	/* find device */
	for (devno = 0; devno < CM_MAX_DEV; devno++) {
		if (dev_table[devno] == link)
			break;
	}
	if (devno == CM_MAX_DEV)
		return;

#ifdef PCMCIA_DEBUG
    printk(KERN_INFO "rs_gpib_detach(0x%p) devno = %d\n", link, devno);
#endif

	rs_gpib_cs_release(link);

	dev_table[devno] = NULL;
	kfree(dev);

	class_device_destroy(gpib_class, MKDEV(major, devno));
    
} /* rs_gpib_detach */


/* Interrupt service routine entry point.
 * 	
 * Arguments:
 * 
 * irq     interrupt number that caused interrupt
 * dev_id  device ID supplied during interrupt registration
 * regs    interrupted processor context
 */
static irqreturn_t rs_gpib_isr(int irq, void *dev_id, struct pt_regs * regs)
{
	struct rs_gpib *dev = (struct rs_gpib *)dev_id;
	int iobase = dev->p_dev->io.BasePort1;

	//++ 080818 - TMS9914 interrupt disable
	outb(0, iobase+IMR0);
	outb(0, iobase+IMR1);
	outb(piimr2, iobase+AUXCR);
	outb(0, iobase+IMR2);
	outb(clrpi, iobase+AUXCR);
	//--

	dev->waitq_complete = 1;
	wake_up_interruptible(&dev->waitq);

	return IRQ_HANDLED;
}

/*======================================================================

    rs_gpib_config() is scheduled to run after a CARD_INSERTION event
    is received, to configure the PCMCIA socket, and to make the
    device available to the system.

======================================================================*/

static void rs_gpib_config(struct pcmcia_device *link)
{
    struct rs_gpib *dev = link->priv;
    tuple_t tuple;
    cisparse_t parse;
    int last_fn = 0;
    int last_ret;
    u_char buf[64];
    config_info_t conf;
    win_req_t req;
    memreq_t map;
    cistpl_cftable_entry_t dflt = { 0 };
    //struct rs_gpib *p = NULL;

    /*
       This reads the card's CONFIG tuple to find its configuration
       registers.
    */
    tuple.DesiredTuple = CISTPL_CONFIG;
    tuple.Attributes = 0;
    tuple.TupleData = buf;
    tuple.TupleDataMax = sizeof(buf);
    tuple.TupleOffset = 0;

	last_ret = pcmcia_get_first_tuple(link, &tuple);
	if(last_ret != CS_SUCCESS) 
	{
		last_fn = GetFirstTuple;
		goto cs_failed;
	}
	last_ret = pcmcia_get_tuple_data(link, &tuple);
	if(last_ret != CS_SUCCESS) 
	{
		last_fn = GetTupleData;
		goto cs_failed;
	}
	last_ret = pcmcia_parse_tuple(link, &tuple, &parse);
	if(last_ret != CS_SUCCESS) 
	{
		last_fn = ParseTuple;
		goto cs_failed;
	}
    link->conf.ConfigBase = parse.config.base;
    link->conf.Present = parse.config.rmask[0];

    /* Look up the current Vcc */
	last_ret = pcmcia_get_configuration_info(link, &conf);
	if(last_ret != CS_SUCCESS) 
	{
		last_fn = GetConfigurationInfo;
		goto cs_failed;
	}
    //link->conf.Vcc = conf.Vcc;

    /*
      In this loop, we scan the CIS for configuration table entries,
      each of which describes a valid card configuration, including
      voltage, IO window, memory window, and interrupt settings.

      We make no assumptions about the card to be configured: we use
      just the information available in the CIS.  In an ideal world,
      this would work for any PCMCIA card, but it requires a complete
      and accurate CIS.  In practice, a driver usually "knows" most of
      these things without consulting the CIS, and most client drivers
      will only use the CIS to fill in implementation-defined details.
    */
    tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;
	last_ret = pcmcia_get_first_tuple(link, &tuple);
	if(last_ret != CS_SUCCESS) 
	{
		last_fn = GetFirstTuple;
		goto cs_failed;
	}
    while (1) 
	{
		cistpl_cftable_entry_t *cfg = &(parse.cftable_entry);
		last_ret = pcmcia_get_tuple_data(link, &tuple);
		if(last_ret != CS_SUCCESS) 
		{
			goto next_entry;
		}
		last_ret = pcmcia_parse_tuple(link, &tuple, &parse);
		if(last_ret != CS_SUCCESS) 
		{
			goto next_entry;
		}
		if (cfg->flags & CISTPL_CFTABLE_DEFAULT) dflt = *cfg;
		if (cfg->index == 0) goto next_entry;
		link->conf.ConfigIndex = cfg->index;

		/* Does this card need audio output? */
		if (cfg->flags & CISTPL_CFTABLE_AUDIO) 
		{
			link->conf.Attributes |= CONF_ENABLE_SPKR;
			link->conf.Status = CCSR_AUDIO_ENA;
		}

		/* Use power settings for Vcc and Vpp if present */
		/*  Note that the CIS values need to be rescaled */
		if (cfg->vcc.present & (1<<CISTPL_POWER_VNOM)) 
		{
			if (conf.Vcc != cfg->vcc.param[CISTPL_POWER_VNOM]/10000)
				goto next_entry;
		}else if (dflt.vcc.present & (1<<CISTPL_POWER_VNOM)) 
		{
			if (conf.Vcc != dflt.vcc.param[CISTPL_POWER_VNOM]/10000)
				goto next_entry;
		}

		if (cfg->vpp1.present & (1<<CISTPL_POWER_VNOM))
			link->conf.Vpp =
				cfg->vpp1.param[CISTPL_POWER_VNOM]/10000;
		else if (dflt.vpp1.present & (1<<CISTPL_POWER_VNOM))
			link->conf.Vpp =
				dflt.vpp1.param[CISTPL_POWER_VNOM]/10000;
	
		/* Do we need to allocate an interrupt? */
		if (cfg->irq.IRQInfo1 || dflt.irq.IRQInfo1)
			link->conf.Attributes |= CONF_ENABLE_IRQ;
	
		/* IO window settings */
		link->io.NumPorts1 = link->io.NumPorts2 = 0;
		if ((cfg->io.nwin > 0) || (dflt.io.nwin > 0)) 
		{
			cistpl_io_t *io = (cfg->io.nwin) ? &cfg->io : &dflt.io;
			link->io.Attributes1 = IO_DATA_PATH_WIDTH_AUTO;
			if (!(io->flags & CISTPL_IO_8BIT))
				link->io.Attributes1 = IO_DATA_PATH_WIDTH_16;
			if (!(io->flags & CISTPL_IO_16BIT))
				link->io.Attributes1 = IO_DATA_PATH_WIDTH_8;
			link->io.IOAddrLines = io->flags & CISTPL_IO_LINES_MASK;
			link->io.BasePort1 = io->win[0].base;
			link->io.NumPorts1 = io->win[0].len;
			if (io->nwin > 1) 
			{
				link->io.Attributes2 = link->io.Attributes1;
				link->io.BasePort2 = io->win[1].base;
				link->io.NumPorts2 = io->win[1].len;
			}
			/* This reserves IO space but doesn't actually enable it */
			last_ret = pcmcia_request_io(link, &link->io);
			if(last_ret != CS_SUCCESS) 
			{
				goto next_entry;
			}
		}
	
		/*
		Now set up a common memory window, if needed.  There is room
		in the dev_link_t structure for one memory window handle,
		but if the base addresses need to be saved, or if multiple
		windows are needed, the info should go in the private data
		structure for this device.
	
		Note that the memory window base is a physical address, and
		needs to be mapped to virtual space with ioremap() before it
		is used.
		*/
		if ((cfg->mem.nwin > 0) || (dflt.mem.nwin > 0)) {
			cistpl_mem_t *mem =
			(cfg->mem.nwin) ? &cfg->mem : &dflt.mem;
			req.Attributes = WIN_DATA_WIDTH_16|WIN_MEMORY_TYPE_CM;
			req.Attributes |= WIN_ENABLE;
			req.Base = mem->win[0].host_addr;
			req.Size = mem->win[0].len;
			if (req.Size < 0x1000)
			req.Size = 0x1000;
			req.AccessSpeed = 0;
			link->win = (window_handle_t)link;
			if(pcmcia_request_window(&link, &req, &link->win) != CS_SUCCESS) goto next_entry;
			map.Page = 0; map.CardOffset = mem->win[0].card_addr;
			if(pcmcia_map_mem_page(link->win, &map) != CS_SUCCESS) goto next_entry;
		}
		/* If we got this far, we're cool! */
		break;
	
		next_entry:
		//if (link->io.NumPorts1)
		//	pcmcia_release_io(link, &link->io);
		last_ret = pcmcia_get_next_tuple(link, &tuple);
		if(last_ret != CS_SUCCESS) 
		{
			last_fn = GetNextTuple;
			goto cs_failed;
		}
	}

    /*
       Allocate an interrupt line.  Note that this does not assign a
       handler to the interrupt, unless the 'Handler' member of the
       irq structure is initialized.
    */
    if (link->conf.Attributes & CONF_ENABLE_IRQ){
	    link->irq.Attributes |= IRQ_HANDLE_PRESENT;
	    link->irq.Handler     = rs_gpib_isr;
	    link->irq.Instance    = link->priv;
		last_ret = pcmcia_request_irq(link, &link->irq);
		if(last_ret != CS_SUCCESS) 
		{
			last_fn = RequestIRQ;
			goto cs_failed;
		}
	}

    /*
       This actually configures the PCMCIA socket -- setting up
       the I/O windows and the interrupt mapping, and putting the
       card and host interface into "Memory and IO" mode.
    */
	last_ret = pcmcia_request_configuration(link, &link->conf);
	if(last_ret != CS_SUCCESS) 
	{
		last_fn = RequestConfiguration;
		goto cs_failed;
	}

    /*
      At this point, the dev_node_t structure(s) need to be
      initialized and arranged in a linked list at link->dev.
    */
    dev->node.major = major;
    dev->node.minor = 0;
    sprintf(dev->node.dev_name, "gpib0");
	dev->node.next = &dev->node;

#ifdef PCMCIA_DEBUG
    /* Finally, report what we've done */
    printk(KERN_INFO "%s: index 0x%02x:",
	   dev->node.dev_name, link->conf.ConfigIndex);
	if (link->conf.Vpp)
		printk(", Vpp %d", link->conf.Vpp);
	if (link->conf.Attributes & CONF_ENABLE_IRQ)
		printk(", irq %d", link->irq.AssignedIRQ);
	if (link->io.NumPorts1)
		printk(", io 0x%04x-0x%04x", link->io.BasePort1,
	       link->io.BasePort1+link->io.NumPorts1-1);
	if (link->io.NumPorts2)
		printk(" & 0x%04x-0x%04x", link->io.BasePort2,
	       link->io.BasePort2+link->io.NumPorts2-1);
	if (link->win)
		printk(", mem 0x%06lx-0x%06lx", req.Base,
	       req.Base+req.Size-1);
    printk("\n");
#endif

    return;

cs_failed:
	cs_error(link, last_fn, last_ret);
    rs_gpib_cs_release(link);

} /* rs_gpib_config */

void rs_gpib_cs_release(struct pcmcia_device *link)
{
#ifdef PCMCIA_DEBUG
	printk(KERN_INFO "rs_gpib_cs_release(0x%p)\n", link);
#endif  
	pcmcia_disable_device(link);
} /* rs_gpib_cs_release */

static struct pcmcia_device_id rs_gpib_ids[] = {
	// # /sbin/pccardctl info
	PCMCIA_DEVICE_MANF_CARD(0x0000, 0x0001),
	PCMCIA_DEVICE_PROD_ID12("RATOC System,Inc.", "REX5052 GPIB PC Card", 0x4031e316, 0x7eedf861),
	PCMCIA_DEVICE_NULL
};
MODULE_DEVICE_TABLE(pcmcia, rs_gpib_ids);

static struct pcmcia_driver rex5052_cs_driver = {
	.owner		= THIS_MODULE,
	.drv		= {
		.name	= "rex5052_cs",
	},
	.probe		= rs_gpib_attach,
	.remove		= rs_gpib_detach,
	.id_table	= rs_gpib_ids,
};




ssize_t gpib_read(struct file *file, char *buffer, size_t count, loff_t *ppos)
{
#ifdef PCMCIA_DEBUG
	printk(KERN_INFO "gpib_read count = 0x%x\n", count);
#endif
	return 0;
}

ssize_t gpib_write(struct file *file, const char *user_buffer, size_t count, loff_t *ppos)
{
#ifdef PCMCIA_DEBUG
	printk(KERN_INFO "gpib_write count = 0x%x\n", count);
#endif
	return 0;
}

int gpib_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	struct rs_gpib *dev = (struct rs_gpib *)file->private_data;
	int iobase = dev->p_dev->io.BasePort1;
	CARD_RESOURCE		card;
	IOPORT_UCHAR		bIOPort;
	IOPORT_USHORT		wIOPort;
	void __user *data;
	int retval = 0;

//#ifdef PCMCIA_DEBUG
//	printk(KERN_INFO "gpib_ioctl cmd = 0x%x\n", cmd);
//#endif
	down(&(dev->lock));
	/* Sanity check to make sure rio is connected, powered, etc */
    if ( dev == NULL)
	{
		retval = -ENODEV;
		goto err_out;
	}

	switch (cmd) {
	case IOCTL_GP_GET_RESOURCE:
		data = (void __user *) arg;
		if (data == NULL)
			break;
		card.IOBaseLow = iobase;
		card.IOBaseHigh = iobase+dev->p_dev->io.NumPorts1-1;
		card.IOLength = dev->p_dev->io.NumPorts1;
		card.IrqLevel = dev->p_dev->irq.AssignedIRQ;
		if (copy_to_user(data, &card, sizeof(CARD_RESOURCE))) {
			retval = -EFAULT;
			goto err_out;
		}
		break;
	case IOCTL_GP_READ_PORT_UCHAR:
		data = (void __user *) arg;
		if (data == NULL)
			break;
		if (copy_from_user(&bIOPort, data, sizeof(IOPORT_UCHAR))) {
			retval = -EFAULT;
			goto err_out;
		}
		bIOPort.Value = inb(iobase+bIOPort.Offset);
		if (copy_to_user(data, &bIOPort, sizeof(IOPORT_UCHAR))) {
			retval = -EFAULT;
			goto err_out;
		}
//#ifdef PCMCIA_DEBUG
//		printk(KERN_INFO "inb: offset = 0x%x value = 0x%x\n", bIOPort.Offset, bIOPort.Value);
//#endif
		break;
	case IOCTL_GP_READ_PORT_USHORT:
		data = (void __user *) arg;
		if (data == NULL)
			break;
		if (copy_from_user(&wIOPort, data, sizeof(IOPORT_USHORT))) {
			retval = -EFAULT;
			goto err_out;
		}
		wIOPort.Value = inw(iobase+wIOPort.Offset);
		if (copy_to_user(data, &wIOPort, sizeof(IOPORT_USHORT))) {
			retval = -EFAULT;
			goto err_out;
		}
#ifdef PCMCIA_DEBUG
		printk(KERN_INFO "inw: offset = 0x%x value = 0x%x\n", wIOPort.Offset, wIOPort.Value);
#endif
		break;
	case IOCTL_GP_WRITE_PORT_UCHAR:
		data = (void __user *) arg;
		if (data == NULL)
			break;
		if (copy_from_user(&bIOPort, data, sizeof(IOPORT_UCHAR))) {
			retval = -EFAULT;
			goto err_out;
		}
		outb(bIOPort.Value, iobase+bIOPort.Offset);
//#ifdef PCMCIA_DEBUG
//		printk(KERN_INFO "outb offset 0x%x value = 0x%x\n", 
//				bIOPort.Offset, bIOPort.Value);
//#endif
		break;
	case IOCTL_GP_WRITE_PORT_USHORT:
		data = (void __user *) arg;
		if (data == NULL)
			break;
		if (copy_from_user(&wIOPort, data, sizeof(IOPORT_USHORT))) {
			retval = -EFAULT;
			goto err_out;
		}
		outw(wIOPort.Value, iobase+wIOPort.Offset);
#ifdef PCMCIA_DEBUG
		printk(KERN_INFO "outw offset 0x%x value = 0x%x\n", 
				wIOPort.Offset, wIOPort.Value);
#endif
		break;
	case IOCTL_GP_START_INTERRUPT:
#ifdef PCMCIA_DEBUG
		printk(KERN_INFO "gpib_ioctl: IOCTL_GP_START_INTERRUPT\n");
		printk(KERN_INFO "gpib_ioctl INTERRUPT...->sleep_on\n");
#endif
		up(&(dev->lock));
		//interruptible_sleep_on(&dev->waitq);
		//++ 080818 - TMS9914 interrupt enable
		outb(0x00, iobase+IMR0);
		outb(0x02, iobase+IMR1);
		outb(piimr2, iobase+AUXCR);
		outb(0xC0, iobase+IMR2);
		outb(clrpi, iobase+AUXCR);
		//--
		dev->waitq_complete = 0;
		wait_event_interruptible(dev->waitq, (dev->waitq_complete) );
		down(&(dev->lock));
#ifdef PCMCIA_DEBUG
		printk(KERN_INFO "gpib_ioctl INTERRUPT...sleep_on->\n");
#endif
		break;
	case IOCTL_GP_STOP_INTERRUPT:
#ifdef PCMCIA_DEBUG
		printk(KERN_INFO "gpib_ioctl: IOCTL_GP_STOP_INTERRUPT\n");
#endif
		//++ 080818 - TMS9914 interrupt disable
		outb(0, iobase+IMR0);
		outb(0, iobase+IMR1);
		outb(piimr2, iobase+AUXCR);
		outb(0, iobase+IMR2);
		outb(clrpi, iobase+AUXCR);
		//--
		dev->waitq_complete = 1;
		wake_up_interruptible(&dev->waitq);
		break;
	default:
		retval = -ENOIOCTLCMD;
		break;
	}

err_out:
	up(&(dev->lock));
	return retval;
}

int gpib_open(struct inode *inode, struct file *file)
{
	struct rs_gpib *dev;
	struct pcmcia_device *link;
	int minor = iminor(inode);

	if (minor >= CM_MAX_DEV)
		return -ENODEV;

	link = dev_table[minor];
	if (link == NULL || !pcmcia_dev_present(link))
		return -ENODEV;

	if (link->open)
		return -EBUSY;

	dev = link->priv;
	file->private_data = dev;

	init_waitqueue_head(&dev->waitq);
	init_MUTEX(&(dev->lock));

	if (file->f_flags & O_NONBLOCK) {
#ifdef PCMCIA_DEBUG
		printk(KERN_INFO "filep->f_flags O_NONBLOCK set\n");
#endif
		return -EAGAIN;
	}

	link->open = 1;

#ifdef PCMCIA_DEBUG
	printk(KERN_INFO "gpib_open version = %s minor = %d\n", version, minor );
	printk(KERN_INFO "irq = %d, io 0x%04x-0x%04x\n", dev->p_dev->irq.AssignedIRQ, 
	dev->p_dev->io.BasePort1, dev->p_dev->io.BasePort1+dev->p_dev->io.NumPorts1-1);
#endif

	return 0;
}

int gpib_release(struct inode *inode, struct file *file)
{
	struct rs_gpib *dev;
	struct pcmcia_device *link;
	int minor = iminor(inode);

#ifdef PCMCIA_DEBUG
	printk(KERN_INFO "gpib_release subminor = %d\n", minor);
#endif
	dev = (struct rs_gpib *)file->private_data;

	if (minor >= CM_MAX_DEV)
		return -ENODEV;

	link = dev_table[minor];
	if (link == NULL)
		return -ENODEV;

	if(link->open == 1){
		link->open = 0;
	}

	return 0;
}

static struct file_operations rex5052_fops = {
	.owner =	THIS_MODULE,
	.read =		gpib_read,
	.write =	gpib_write,
	.ioctl =	gpib_ioctl,
	.open =		gpib_open,
	.release =	gpib_release,
};

static int __init rex5052_init_module(void)
{
	int rc;

	gpib_class = class_create(THIS_MODULE, "rex5052_gpib");
	if (!gpib_class)
		return -1;

	/* register this driver*/
	major = register_chrdev(0, "gpib", &rex5052_fops);
	if (major < 0) {
#ifdef PCMCIA_DEBUG
		printk(KERN_INFO "register_chrdev failed. Error number %d", major);
#endif
	}
#ifdef PCMCIA_DEBUG
	printk(KERN_INFO "rex5052_init_module: loading... major %d\n", major);
#endif

	rc = pcmcia_register_driver(&rex5052_cs_driver);
	if (rc < 0) {
		unregister_chrdev(major, "gpib");
		return rc;
	}

	return 0;
}

static void __exit rex5052_exit_module(void)
{
#ifdef PCMCIA_DEBUG
	printk(KERN_INFO "rex5052_exit_module: unloading\n");
#endif

	/* unregister this driver */
	pcmcia_unregister_driver(&rex5052_cs_driver);
	unregister_chrdev(major, "gpib");
	class_destroy(gpib_class);
}

module_init (rex5052_init_module);
module_exit (rex5052_exit_module);
MODULE_LICENSE("Dual BSD/GPL");

